/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mvif - mobile virtual interface 
 *
 * Overview: -   The mvif device driver is a virtual device.  It is
 * essentially a routing tool.
 *
 * It does unicast tunneling i/o for unicast IP packets.  This is intended
 * to be used with Mobile IP at a Home Agent to tunnel packets to
 * a remote COA (Care Of Address), typically a Foreign Agent,
 * where the packets will be decapsulated
 * and delivered locally.  A Mobile Node might use this facility to tunnel
 * back to its home network via the HA.  Tunneling is done via net protocol 4 
 * IPIP which is the same as for mrouted tunnels.  
 * This driver coexists with mrouting functionality due to the mvifinput
 * switch.  If set to 1, unicast input ipip packets are sent here, else
 * they would be fed upstairs to a raw socket or discarded.
 *
 * This interface is "dynamic" in that routes to hosts and nets 
 * may be arbitrarily inserted into and removed from the routing table with: 
 *
 * # route add -tunnel remote_host tunnel_dst_address(COA) 
 *
 * as a result a routing daemon for Mobile-IP can insert and remove
 * tunnel routes from the HA to COA address.  There is no need
 * for a MIP "visitor list" as state (MN/COA) can be kept in the routing
 * table on a per route basis.  Thus this driver allows Mobile IP
 * to use and take advantage of the built-in BSD routing radix algorithm.
 *
 * kernel config:
 *	To include the driver, the kernel config file must have:
 *
 *	options   MIPTUNNEL
 *	pseudo-device   mvif    1
 *
 * setup:
 *	In order to send packets,  the mvif driver must have an IP
 *	address bound to it (for routing table reasons).  Unfortunately
 *	this makes no sense whatsoever as the driver is a virtual
 *	routing function and one does not ordinarily write
 *	packets directly to this "interface".  Packets directly written
 *	to it are rejected immediately with a host unreachable error.
 *
 *	For normal operation, we bind one of the "local-only" IP 
 *	network addresses to it.   E.g., class A
 *	net 10.0.0.0, and there are class B/C too.  These are IP
 *	addresses that cannot be used in Internet routing.  You might use:
 *	(See /etc/hosts or RFC1597.  192.168.2.2 is one such class C address).
 *
 *	# ifconfig mvif0 inet 192.168.2.2
 *
 *	on all such interfaces.
 *
 *	A HA needs a bound address so that tunnel routes will work;
 *	i.e., the binding need only take place at a tunnel entry point.
 *	Other Mobile-IP systems (FAs/Mobile Nodes) do not need such
 *	a binding, as they only accept tunnel packets on input.
 *
 *	The above config can be done in /etc/sysconfig, of course.
 *	Therefore for a HA function or wherever route -tunnel
 *	is used, you must bind a (useless) IP address.
 *	For FA functionality (to be a tunnel endpoint), there is no need.
 *
 *	One *might* choose to bind a routable IP address to this interface.
 *	One should however for normal operation make sure that the 
 *	inner ip src is set to a real IP address for the router in question.  
 * 	This can be done via the mvifconfig app see below.  The driver
 *	supports 2 ioctls for setting various ip src addresses.
 *
 *	Thus normal setup might be:
 *	# ifconfig mvif0 inet 192.168.2.2
 *	# mvifconfig -w -i  200.1.2.3  
 *
 *	where 200.1.2.3 is say an ip address bound to an ethernet if
 *	on this box.  192.168.2.2 will NOT appear in the packet as an ip
 *	src address.
 *
 *	Packets sent into a routing tunnel have an ip header typically
 *	with an ip src and ip dst set.  The output function slaps on
 *	another ip header, where the ip dst is the tunnel endpoint (COA),
 *	and the src is left alone and will be filled in according to
 *	the real interface that the packet is sent out.  Thus a packet
 *	is routed twice (through the tunnel, and say perhaps through a
 *	default route).   It will have an OUTER IP header and an INNER
 *	IP header.  When it arrives, the OUTER ip header is stripped
 *	and the remaining ip packet is forwarded according to normal
 *	ip routing.
 *
 * output routine: mvifoutput
 *
 *	prepends IP header with ip dst to to gateway (COA).
 *	calls ip_output with FORWARDING set, thus routing will take place.
 *
 *	Consider this case:
 *	We want to send pkts to MN X at FA Y.  Our routing table has in
 *	it:
 *
 *	To default via local_router bound to Ethernet device
 *	To MN  via  COA(FA IP)  bound to tunnel device.
 * 
 *	Therefore a packet routed to the tunnel route will be delivered
 *	to ip_output and routed according to the COA.  
 *	Then with a new ip outer header (ip_p == IPIP), the
 *	packet will be given to ip_output and routed according
 *	to the FA's ip address (COA).  It might then be
 *	routed out the default route on a real device.
 *
 * input routine:  mvif_input
 *
 *	called from ipintr (IP input) (eventually as it pays a visit
 *	to multicast input first (MROUTING) which may or may not be compiled
 *	on).  If MROUTING doesn't want the packet (it's unicast on
 *	the inside, not multicast), it arrives here.
 *	Once a packet arrives note that the outer
 *	ip header has ip_p set to IPIP.  The job of this routine is to strip
 *	off the outer header and deliver the remaining packet to
 *	ipintr, where it will be routed normally.
 *
 * security analysis:
 *	Many border routers have filters set that drop
 *	packets from the outside with an ip_src that matches subnets
 *	on the inside.  It is thought that this provides some sort
 *	of spoofing protection.  An IPIP input function (such as with
 *	a Mobile-IP FA) punches a hole through that filter.
 *
 *	The variable mvifinput may be turned off via sysctl
 *	(which must be recompiled) at a HA to prevent IPIP
 *	input.  It cannot be turned off at a FA.  We suggest
 *	that IPSEC  AH (authentication is needed) so that tunnels
 *	can be made secure from end to end and this driver should only
 *	allow authenticated packets through.
 *
 *	If IPSEC is used and available, the variable mvifipsecinput
 *	may be turned on as appropriate to force IPIP input packets
 *	to have an AH or ESP mbuf attribute; i.e., they must already
 *	have been processed on this computer by AH/ESP or they WILL
 *	be processed on this computer as the AH/ESP headers are 
 *	still upstairs and not processed yet.  This means
 *	that the endpoint of a tunnel can be made secure via AH/ESP.
 *	This implies a point to point connection between the tunnel
 *	src and tunnel destination.
 *
 * ioctls:
 *	The mvifconfig routine may be used to set both the inner ip
 *	src address and the outer ip src address.  The driver has a tricky
 *	problem, which is that if a packet has been handed to the output
 *	routine and the packet originated on this system (not forwarded),
 *	the kernel has to find out that getting an ip address from the
 *	the mvif device is wrong.  Therefore the device needs to
 *	be configured with an additional ip addr that is one for a
 *	a real device (say ethernet) with a real IP address binding.
 *	In order to support the use of non-routable IP addresses bound
 *	to this interface,  the driver knows its own ip address and
 *	throws packets away if they show up with its ip address.  
 *	It is necessary to configure
 *	the driver with a "real" but borrowed ip address from another
 *	real interface, else packets to the tunnel originating on
 *	this system will not survive.  
 *
 *	It is also possible as a feature to set the IP src for the outer
 *	packet.  Thus it might be possible for a MN to use a FA's COA
 *	for packets it wishes to tunnel back to the HA.  Or a secure
 *	tunnel (IPSEC) may be setup since the outer ip src must be 
 *	found in the IPSEC uni-directional SA (spi, ip src, ip dst)
 *	on both this computer and this tunnel endpoint.
 *
 *	IP addresses for inner/outer src values may be bound as follows:
 *
 *	mvifconfig -w -i <inner ip>
 *	mvifconfig -w -o <outer ip>
 *
 *	Note that the ip address must be an ip address, not a DNS hostname.
 *	The inner binding must be done AFTER an ifconfig for a real
 *	device.  The outer binding need not be done for normal operation.
 */
#include "mvif.h"
#if NMVIF > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/syslog.h>
#include <sys/proc.h>
#include <machine/cpu.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>
#include <net/bpf.h>

#ifdef	INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#endif

#ifdef IPSEC
#include <netkey/key.h>
#include <netsec/ipsec.h>
#include <netsec/ipsec_ah.h>
#include <netsec/ipsec_esp.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#ifdef ISO
#include <netiso/iso.h>
#include <netiso/iso_var.h>
#endif

#include <netinet/if_mvif.h>

#include "bpfilter.h"

#define MVIFMTU	16384
#define	satosin(sa)	((struct sockaddr_in *)(sa))
#define IP_HDR_LEN 20

/* ip address of interface */
static struct in_addr mvif_sa;

/* inner ip address.  what we set inner ip on IPIP output to 
 * note: 0 means OFF
 */
static struct in_addr inip;
static struct ifaddr *inner_ifa;   /* ifa for inip */

/* outer ip address.  what we set outer ip on IPIP output to 
 * note: 0 means OFF
 */
static struct in_addr outip;

int mvifdebug = 0;  /* turns on debug */

/* input filters - controllable via systl  ********************/

int mvifinput = 1;  /* sysctl: controls whether IPIP pkts accepted or not */

/* sysctl: only pkts with IPSEC attributes from prior AH/ESP decode
 * are accepted as input
 */
int mvifipsecinput = 0;  

int mvifipsecspoofs = 0;	/* counter of tossed pkts if ipsecinput set */

#ifdef IDEBUG
extern int idebug;
#endif

static void
mvifattach(void *dummy)
{
	register struct ifnet *ifp;
	register int i = 0;

	
	printf("mvif0: mobile virtual interface/tunnel device\n");
	ifp = &mvif;
	ifp->if_name = "mvif";
        ifp->if_next = NULL;
	ifp->if_unit = i;
	ifp->if_mtu = MVIFMTU;
	ifp->if_flags = IFF_NOARP;
	ifp->if_ioctl = mvifioctl;
	ifp->if_output = mvifoutput;
	ifp->if_type = IFT_OTHER;
	ifp->if_hdrlen = 0;
	ifp->if_addrlen = 0;
	if_attach(ifp);

	inip.s_addr = outip.s_addr = 0;
}

PSEUDO_SET(mvifattach, if_mvif);

/*
 * mvifoutput()
 *
 * called to encapsulate ip packet.  This routine
 * will append an IPIP header on the packet.
 * The gateway in the route contains the COA.
 *
 * The logic here is that packets routed directly
 * to the tunnel route or forwarded to it 
 * end up here, a new ip header is put on,
 * the ip dst is made the COA (the route's gateway),
 * and the packet is handed back to ip_output.
 *
 * The driver does not permit packets to be sent directly
 * on any non-tunnel route; i.e., the associated route
 * must have RTF_TUNNEL set.   If you bind an ip address
 * to the device (with route flags set to Uc say)
 * and try to send packets directly to it,
 * they will be tossed.
 *
 * overview:
 *	add an ip header
 *	ip_dst is dst passed in
 *	ip_src is 0, so ip_output will choose it acc. to true i/f
 *	ip_proto is IPPROTO_IPIP (4)
 *	deliver to ip_output with no route (we want it to route the pkt)
 */

#define ENCAP_TTL 64
#define ENCAP_PROTO IPPROTO_IPIP        /* 4 */
    
/* prototype IP hdr for encapsulated packets */
struct ip unicast_encap_iphdr = {
#if BYTE_ORDER == LITTLE_ENDIAN
        sizeof(struct ip) >> 2, IPVERSION,
#else
        IPVERSION, sizeof(struct ip) >> 2,
#endif
        0,                              /* tos */
        sizeof(struct ip),              /* total length */
        0,                              /* id */
        0,                              /* frag offset */
        ENCAP_TTL, ENCAP_PROTO,
        0,                              /* checksum */
};
 
/*
 * note: ip len, off, id have been byte-swapped on inner packet
 */

int
mvifoutput(ifp, m, dst, rt)
	struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;	/* gateway in routing table */
	register struct rtentry *rt;
{
	register struct ip *ip = mtod(m, struct ip *); 
        register struct ip *ip_copy;
        int hlen = ip->ip_hl << 2;   	/* hlen - length of ip header bytes */
	register int len = ntohs(ip->ip_len);  /* len of packetn bytes  */
	int errval = 0;

	/* New for 2.2: ip_output must have a route.  This code is
	 * based on vaguely similar treachery in tcp_subr.c.
	 * There are several gains in making this static:
	 *	1. nextroute need not be initialized every time through.
	 *	2. Its ro_rt need not be free'd.
	 *	3. It functions as a (one-entry) route cache for tunnels.
	 * Whoopee.
	 */
	static struct route nextroute;

	/* incantations needed to get at ip address
	 *		satosin(rt->rt_gateway)->sin_addr.s_addr 
	 *		satosin(dst)->sin_addr.s_addr
	 */

	/* paranoid checks
	*/
	if (!rt) {
		ifp->if_oerrors++;
		log(LOG_WARNING,"mvif: rt is NULL, possible routing table problem\n"); 
		m_freem(m);
		return(EHOSTUNREACH);
	}
	if (rt->rt_gateway->sa_family != AF_INET) {
		panic("mvif: wrong address family");
	}
	
	if (mvifdebug) {
		printf("mvifoutput called ip outer dst %x, coa %x len %x m %x, (inner) ip src %x\n",
 	 		satosin(dst)->sin_addr.s_addr, 
  			satosin(rt->rt_gateway)->sin_addr.s_addr, len,
			m,
			ip->ip_src 
			);
	}

	/* error check.  if route not tunnel route OR dst is somehow
	 * this device's ip address, toss the packet.
	 * This rules out writing to the mvif device
	 * when an ip address is bound to it.
	 */
	if (((rt->rt_flags & RTF_TUNNEL) == 0) || 
		(mvif_sa.s_addr == satosin(dst)->sin_addr.s_addr)) {
		ifp->if_oerrors++;
		m_freem(m);
		return(EHOSTUNREACH);
	}

	/* rule out recursive use of device.  We mark this packet
	 * as having gone through the device once.   Once is enough.
	 */
	if ( m->m_flags & M_TUNNEL ) {
		ifp->if_oerrors++;
		log(LOG_WARNING,"mvif: recursive use of tunnel device: possible routing table problem\n"); 
		m_freem(m);
		return(EHOSTUNREACH);
	}
	m->m_flags |= M_TUNNEL;

#ifdef IPSEC
	if (rt->rt_seca[RTSA_AHTUNNEL]) {
		ipsec_ipstat.ipsec_route_ah_tunnel_needed++;
	}
	if (rt->rt_seca[RTSA_ESPTUNNEL]) {
		ipsec_ipstat.ipsec_route_esp_tunnel_needed++;
	}
#endif

	m = mvif_prependip(
		ifp, 
		m, 
		satosin(dst)->sin_addr,
		outip,
		&errval,
		len    /* no checksum as ip_output will do it */
#ifdef IPSEC
		,
		rt->rt_seca[RTSA_AHTUNNEL],
		rt->rt_seca[RTSA_ESPTUNNEL]
#endif
		);

	if (m == NULL) {
		return(errval);
	}

	ip_output(m, (struct mbuf *)0, &nextroute,
		  IP_FORWARDING, (struct ip_moptions *)0
#ifdef IPSEC
		  ,0
#endif
		);
	return(0);
}

/* 
 * unicast tunnel input routine. 
 * basic idea:
 *	called from ip_intr for IPIP unicast packets
 *	if MROUTING is defined, we pay a short visit to
 *		X_ipip_input and then come here else
 *	we visit
 *		ipip_input and come here
 *		
 *	throw out ip options if any
 *	throw away ipip header
 *	deliver inside packet to ip_intr
 *		which will deliver locally or forward
 *		acc. to routes in routing table.
 *	Note: that no forwarding will occur if
 *		there is not a route to the destination.
 */

void
mvif_input(m, hlen)
struct mbuf *m;
int hlen;	/* ip already calculated header length */
{
	register struct ip *ip;
	register struct ifqueue *ifq;
	register int s;
	register struct in_ifaddr *ia;
	struct ifnet *ifp = &mvif;
	int isr = NETISR_IP;

	/* allow no packets through if set to 0
	*/
	if (mvifinput == 0) {
		m_freem(m);
		return;
	}


	if (mvifdebug) {
		printf("mvif_input hlen %x\n", hlen);
	}

	/* throw out any options, although ip itself deals
	 * with them.
	 */
	if (hlen > IP_HDR_LEN) {
		ip_stripoptions(m, (struct mbuf *) 0);
		hlen = sizeof(struct ip);
	}
	
	ip = mtod(m, struct ip *);
	if (m->m_len < hlen + sizeof (struct ip)) {
		if ((m = m_pullup(m, hlen + sizeof (struct ip))) == 0) {
			/*
			mvifdrops++;
			*/
			return;
		}
		ip = mtod(m, struct ip *);
	}

	/* throw out the IPIP header
	*/
	m->m_data += IP_HDR_LEN;
	m->m_len -= IP_HDR_LEN;
	m->m_pkthdr.len -= IP_HDR_LEN;
	m->m_pkthdr.rcvif = ifp;

#ifdef IPSEC
	/* if set to 1, only allow packets with IPSEC attributes
	 * that must be shared with this router; i.e., we have
	 * a key relationship with the sender and do the following:
	 *	1. esp/ah already successfully processed, then ok
	 *	2. packet was of form IP|IP|IPSEC (AH|ESP)
	 *	and inner header has dest this box (promise that we
	 *	will do IPSEC soon on this system), then ok
	 *	3. else throw away.
	 */
	if (mvifipsecinput) {
		if ( m->m_flags & (M_DECRYPTED | M_AUTHENTIC) ) {
			goto safe;
		}
		else {
			/* form pointer to inner ip header
			*/
			ip = mtod(m, struct ip *);
			hlen = ip->ip_hl << 2;
			if (m->m_len < hlen + sizeof (struct ip)) {
				if ((m = m_pullup(m, hlen + sizeof (struct ip))) == 0) {
					return;
				}
				ip = mtod(m, struct ip *);
			}
			/* packet should be addressed to ME, and
			 * next header should be IPSEC header
			 */
			if (ip->ip_p == IPPROTO_AH || ip->ip_p == IPPROTO_ESP) {
				for (ia = in_ifaddr; ia; ia = ia->ia_next) {
					if (IA_SIN(ia)->sin_addr.s_addr 
							== ip->ip_dst.s_addr) {
						goto safe;
					}
				}
			}
		}
		/* else throw packet away */
		mvifipsecspoofs++;
		m_freem(m);
		return;
	}
safe:
#endif
	if (mvifdebug) {
		ip = mtod(m, struct ip *);
		printf("mvif_input(): inner ip_src %x ip_dst %x\n", ip->ip_src, ip->ip_dst);
	}

	/* deliver to ip input function
	*/
	ifq = &ipintrq;
	s = splimp();
	if (IF_QFULL(ifq)) {
		IF_DROP(ifq);
		m_freem(m);
	} else {
		IF_ENQUEUE(ifq, m);
		schednetisr(isr);
	}
	/* count packets against mvif interface
	*/
	ifp->if_ipackets++;
	/*
	ifp->if_ibytes += m->m_pkthdr.len;
	*/
	splx(s);
}


/* ARGSUSED */
void
mvifrtrequest(cmd, rt, sa)
	int cmd;
	struct rtentry *rt;
	struct sockaddr *sa;
{

	if (rt)
		rt->rt_rmx.rmx_mtu = MVIFMTU;
}

/*
 * Process an ioctl request.
 */
/* ARGSUSED */
int
mvifioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	int cmd;
	caddr_t data;
{
	register struct ifaddr *ifa;
	register struct ifreq *ifr = (struct ifreq *)data;
	register int error = 0;
	struct proc   *p = curproc; /* XXX */

	switch (cmd) {

	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;
		ifa = (struct ifaddr *)data;
		if (ifa != 0 && ifa->ifa_addr->sa_family == AF_ISO)
			ifa->ifa_rtrequest = mvifrtrequest;
		/*
		 * Everything else is done at a higher level.
		 */

		/* store the device ip address in a handier place
		*/
		mvif_sa = ((struct sockaddr_in *) ifa->ifa_addr)->sin_addr;
		if (mvifdebug)
			printf("mvif ip: %lx\n", mvif_sa.s_addr);
		break;

	/* yeah, right.  Irony here
	*/
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		if (ifr == 0) {
			error = EAFNOSUPPORT;		/* XXX */
			break;
		}
		switch (ifr->ifr_addr.sa_family) {

#ifdef INET
		case AF_INET:
			break;
#endif

		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	case SIOCSIFMTU:
		ifp->if_mtu = ifr->ifr_mtu;
		break;

	/* get inner ip address */
	case SIOCGMVINIP:     
		ifr->ifr_data = (caddr_t) inip.s_addr;
		break;

	/* set inner ip address */
	case SIOCSMVINIP:  
 		error = suser(p->p_ucred, &p->p_acflag);
                if (error != 0) {
                        break;
                }

		/* note: 0 means this facility is turned off
		*/
		inip.s_addr = (u_long) ifr->ifr_data;
		if (mvifdebug) {
			printf("inip %lx\n", inip.s_addr);
		}
		mvif_getifa();
		if (inner_ifa == NULL) {
			error = EINVAL;
		}
		break;

	/* get outer ip address */
	case SIOCGMVOUTIP: 
		ifr->ifr_data = (caddr_t) outip.s_addr;
		break;

	/* set outer ip address */
	case SIOCSMVOUTIP:   
 		error = suser(p->p_ucred, &p->p_acflag);
                if (error != 0) {
                        break;
                }
		/* note: 0 means this facility is turned off
		*/
		outip.s_addr = (u_long) ifr->ifr_data;
		if (mvifdebug) {
			printf("outip %lx\n", outip.s_addr);
		}
		break;

	default:
		error = EINVAL;
	}
	return (error);
}

/*
 * function needed for -tunnel code in route.c
 *
 * return associated ifaddr pointer.
 */
struct ifaddr *
ifa_mip_tunnel(void)
{
        struct ifnet *ifp = &mvif;
        struct ifaddr *ifa = 0;         
        register struct in_ifaddr *ia = 0;

        IFP_TO_IA(ifp, ia);
        if (ia)
                ifa = &(ia->ia_ifa);
        return(ifa);
}

/*
 * map ip address to interface address pointer
 */
static
struct  sockaddr_in sa_in = { sizeof(struct sockaddr_in), AF_INET };

static void
mvif_getifa(void)
{
	struct ifaddr *ifa;
	struct sockaddr_in *sin;
	struct ifnet *ifp;

	sa_in.sin_addr.s_addr = inip.s_addr;
	sa_in.sin_port = htonl(0);

	inner_ifa = ifa_ifwithaddr((struct sockaddr *) &sa_in);	

}

/*
 * return indirect (acc to inner ip)
 * address pointer.  This is so that
 * routing functions that need to point
 * to a real device can do so
 */
struct in_ifaddr *
mvif_requestia(void)
{
	return((struct in_ifaddr *) inner_ifa);
}

struct in_addr
mvif_getipaddr(void)
{
	return(inip);
}

/*
 * mvif_prependip - slap on an outer ip header
 *	ipsec routing is done here.
 *
 * assumption:
 *	 on error, everything here frees the mbuf chain.
 * returns:
 *	mbuf chain. NULL on error.
 *	*errval - returns errno (initialized to 0 here).
 *
 *	0 - no error
 *	> 1: error value
 *		ENOBUFS - ran out of mbufs
 *		EIPSEC - ipsec error
 */

struct mbuf *
mvif_prependip(
	struct ifnet *ifp,	/* if pointer, may be NULL */
	struct mbuf *m,		/* the mbuf chain for the packet */
	struct in_addr outdst,	/* outer ip dst */
	struct in_addr outsrc,  /* if non-zero, outer ip src addr is set */
	int *errval,  		/* errno if found */
	int len			/* length in bytes of header */
#ifdef IPSEC
	,
	struct key_tblnode *ahsa,
	struct key_tblnode *espsa
#endif
)
{
	register struct ip *ip = mtod(m, struct ip *); 
        register struct ip *ip_copy;

	*errval = 0;
	/* if m_prepend fails, the mbuf chain is freed there,
	 * so we don't do it here.
	 */
	M_PREPEND(m, sizeof(struct ip), M_DONTWAIT);
	if (m == 0) {
		if (ifp)
			ifp->if_oerrors++;
		*errval = ENOBUFS;
		return(NULL);
	}
	/*
	 * fill in the encapsulating IP header.
	 */
	ip_copy = mtod(m, struct ip *);
	*ip_copy = unicast_encap_iphdr;
	ip_copy->ip_id = htons(ip_id++);
	ip_copy->ip_len += len;
	if (outsrc.s_addr) {
		ip_copy->ip_src.s_addr = outsrc.s_addr; 
		if (mvifdebug) {
			printf("mvif: outer ip src %x\n", 
				ip_copy->ip_src.s_addr);
		}
	}
	/* force ip to fill this in */
	else {
		ip_copy->ip_src.s_addr = INADDR_ANY;
	}
	ip_copy->ip_dst.s_addr = outdst.s_addr;  
	if (mvifdebug) {
		printf("mvif: outer ip dst %x\n", ip_copy->ip_dst);
    	}

	if (ifp)
		ifp->if_opackets++;

#ifdef IPSEC
	/* if ip_copy->ip_src is zero, then write in the src of the 
	 * SA (pair).  then check that ip_copy src/dst is equal to
	 * SA (pair) src/dst.  if either check fails, log the result
	 * and drop the packet.
	 */
	if (ahsa || espsa) {
		struct key_tblnode *ansa = (ahsa) ? ahsa : espsa;
		if (ip_copy->ip_src.s_addr == INADDR_ANY) {	/* write src */
			ip_copy->ip_src.s_addr = 
				 satosin(ansa->secassoc->src)->sin_addr.s_addr;
		}
		else if (ip_copy->ip_src.s_addr !=	/* check src */
			satosin(ansa->secassoc->src)->sin_addr.s_addr) {
                        *errval = EIPSEC;
		}

		if (ip_copy->ip_dst.s_addr !=		/* check dst */
			satosin(ansa->secassoc->dst)->sin_addr.s_addr) {
                        *errval = EIPSEC;
		}

		if (*errval) {
			printf("mvif_prependip(): src or dst of an outer header does not match SA src/dst pair: \n");
			printf("mvif_prependip(): src/dst for IP %s/",
				inet_ntoa(ip_copy->ip_src));
			printf("%s, ",
				inet_ntoa(ip_copy->ip_dst));
			printf("for SA %s/", inet_ntoa(
				satosin(ansa->secassoc->src)->sin_addr));
			printf("%s.\n", inet_ntoa(
				satosin(ansa->secassoc->dst)->sin_addr));

			/* TBD: log this in ipsec_ipstats? */
                        m_freem(m);
                        return(NULL);
		}
	}

	/* if esp flag, we shove esp header
	 * in between outer and inner ip headers
	 */
	if (espsa) {
		int status = 0;
		ipsec_ipstat.ipsec_route_esp_needed++;

		status = ipsec_esp_transport_output(m, espsa->secassoc);
		ip_copy = mtod(m, struct ip *);
		ip_copy->ip_len = m->m_pkthdr.len;
		if (status < 0) {
			if (ifp)
				ifp->if_oerrors++;
			ipsec_ipstat.ipsec_error++;
			ipsec_ipstat.ipsec_route_esp_failed++;
			*errval = EIPSEC;
			return(NULL);
		}
	}

	/* if ah flag indicated, we shove in ipsec AH between
	 * outer IP and inner IP.  result is:
	 * outer IP (ip_p = AH)
	 * AH header (ip_p = IPIP)
	 * inner IP
	 * note: for this to work the outer ip needs to
	 * be bound to a real i/f (e.g., ethernet)
	 */
	if (ahsa) {
		ipsec_ipstat.ipsec_route_ah_needed++;

		m = ipsec_ah_output(m, ahsa->secassoc);
		if (!m) {
			if (ifp)
				ifp->if_oerrors++;
			ipsec_ipstat.ipsec_error++;
			ipsec_ipstat.ipsec_route_ah_failed++;
			*errval = EIPSEC;
			return(NULL);
		}
		ip_copy = mtod(m, struct ip *);
		ip_copy->ip_len = m->m_pkthdr.len;
	}
#endif /* IPSEC */

	return(m);
}

#endif /* MIPVIF > 0 */


